#include "eglwrapper.h"

#include <assert.h>
#include <GLES2/gl2.h>
#include <EGL/egl.h>

namespace SuperScaler {

#define BACKEND_RPI
#define arraysize(X) (sizeof(X) / sizeof((X)[0]))

class EGLWindow : public GLWindowInterface {
public:
  void Setup(EGLNativeDisplayType native_display,
	     EGLNativeWindowType native_window) {
    egl_display = eglGetDisplay(native_display);
    assert(eglInitialize(egl_display, NULL, NULL));
    const EGLint attrs[] = {
      EGL_BLUE_SIZE, 5,
      EGL_GREEN_SIZE, 5,
      EGL_RED_SIZE, 5,
      EGL_NONE,
    };
    const EGLint contextAttrs[] = {
      EGL_CONTEXT_CLIENT_VERSION, 2,
      EGL_NONE, EGL_NONE
    };

    EGLConfig config;
    int num_configs;
    assert(eglChooseConfig(egl_display, attrs, &config, 1, &num_configs));
    assert(num_configs == 1);

    egl_surface = eglCreateWindowSurface(egl_display,
					 config,
					 native_window,
					 NULL);
    egl_context = eglCreateContext(egl_display,
				   config,
				   EGL_NO_CONTEXT,
				   contextAttrs);
    assert(eglMakeCurrent(egl_display,
			  egl_surface,
			  egl_surface,
			  egl_context));
  }

  ~EGLWindow() {
    eglDestroyContext(egl_display, egl_context);
    eglDestroySurface(egl_display, egl_surface);
    eglTerminate(egl_display);
  }

  virtual void SwapBuffers() {
    eglSwapBuffers(egl_display, egl_surface);
  }

  EGLDisplay egl_display;
  EGLSurface egl_surface;
  EGLContext egl_context;
};


#ifdef BACKEND_RPI
#include <bcm_host.h>

class EGLWindowRPI : public EGLWindow {
public:
  EGLWindowRPI() {
    bcm_host_init();

    DISPMANX_ELEMENT_HANDLE_T dispman_element;
    DISPMANX_DISPLAY_HANDLE_T dispman_display;
    DISPMANX_UPDATE_HANDLE_T dispman_update;
    VC_RECT_T dst_rect;
    VC_RECT_T src_rect;
   
    // create an EGL window surface, passing context width/height
    assert(graphics_get_display_size(0 /* LCD */, &display_width, &display_height) >= 0);

    dst_rect.x = 0;
    dst_rect.y = 0;
    dst_rect.width = display_width;
    dst_rect.height = display_height;
      
    src_rect.x = 0;
    src_rect.y = 0;
    src_rect.width = display_width << 16;
    src_rect.height = display_height << 16;   

    dispman_display = vc_dispmanx_display_open( 0 /* LCD */);
    dispman_update = vc_dispmanx_update_start( 0 );
    VC_DISPMANX_ALPHA_T alpha;
    alpha.flags = DISPMANX_FLAGS_ALPHA_FIXED_ALL_PIXELS;
    alpha.opacity = 0xffu;
    alpha.mask = 0;

    dispman_element = vc_dispmanx_element_add (
      dispman_update,
      dispman_display,
      0,  /*layer*/
      &dst_rect,
      0,  /*src*/
      &src_rect,
      DISPMANX_PROTECTION_NONE,
      &alpha,
      0,  /*clamp*/
      DISPMANX_NO_ROTATE);
      
    nativewindow.element = dispman_element;
    nativewindow.width = display_width;
    nativewindow.height = display_height;
    vc_dispmanx_update_submit_sync( dispman_update );

    Setup(EGL_DEFAULT_DISPLAY, &nativewindow);
  }

  virtual int xsize() { return display_width; }
  virtual int ysize() { return display_height; }

  uint32_t display_width;
  uint32_t display_height;
  EGL_DISPMANX_WINDOW_T nativewindow;
};

typedef EGLWindowRPI EGLBackend;

#endif

#ifdef BACKEND_X
#include <X11/Xlib.h>
#include <X11/Xutil.h>

class EGLWindowX : public EGLWindow {
public:
  EGLWindowX() {
    Window root;
    XSetWindowAttributes swa;
    XSetWindowAttributes  xattr;
    XWMHints hints;
    Window win;

    x_display = XOpenDisplay(NULL);
    assert(x_display != NULL);

    root = DefaultRootWindow(x_display);

    swa.event_mask  =  ExposureMask | PointerMotionMask | KeyPressMask;
    win = XCreateWindow(
               x_display, root,
               0, 0, esContext->width, esContext->height, 0,
               CopyFromParent, InputOutput,
               CopyFromParent, CWEventMask,
               &swa );

    xattr.override_redirect = FALSE;
    XChangeWindowAttributes ( x_display, win, CWOverrideRedirect, &xattr );

    hints.input = TRUE;
    hints.flags = InputHint;
    XSetWMHints(x_display, win, &hints);

    // make the window visible on the screen
    XMapWindow (x_display, win);
    XStoreName (x_display, win, title);

    Setup(x_display, (EGLNativeWindowType) win);
  }

  XDisplay x_display;
};

typedef EGLWindowX EGLBackend;
#endif


GLWindowInterface *create_gl_window() {
  return new EGLBackend();
}
}  // superscaler
