#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include <assert.h>

#include <map>
#include <queue>
#include <set>
#include <vector>

#include "picdb.h"
#include "exif.h"

using namespace std;

string db_dir = "MUST SET --db_dir=";

pthread_mutex_t global_filename_lock = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t history_lock = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t db_lock = PTHREAD_MUTEX_INITIALIZER;
refptr<ImageInterface> quickmemsave;

class Locker {
public:
  Locker(pthread_mutex_t *l) : lock_(l) {
    pthread_mutex_lock(lock_);
  }
  ~Locker() {
    pthread_mutex_unlock(lock_);
  }
private:
  pthread_mutex_t *lock_;
};

string readstring(FILE *f) {
  string ret;
  while (true) {
    int c = fgetc(f);
    if (c == EOF || c == '\n' || c=='=') {
      ungetc(c, f);
      break;
    }
    if (c == '\\') c = fgetc(f);
    ret.push_back(c);
  }
  return ret;
}

bool writestring(FILE *f, const string& str) {
  for (int i = 0; i < str.size(); i++) {
    if (str[i] == '\n' || str[i] == '=') {
      if (fputc('\\', f) == EOF) return false;
    }
    if (fputc(str[i], f) == EOF) return false;
  }
  return true;
}

string picdb_name(const Filename &file) {
  string p = file.path();
  string ret = db_dir;
  for (int i = 0; i < p.size(); i++) {
    switch (p[i]) {
      case '/': ret.append("_s_"); break;
      case '_': ret.append("__"); break;
      default: ret.push_back(p[i]);
    }
  }
  return ret;
}

int resolve(int old, int from_file, int start_value) {
  if (old == start_value) return from_file;
  if (from_file == start_value) return old;
  if (start_value == 1 && old > 1 && from_file > 1) return max(old, from_file);
  return from_file;
}

class Image : public ImageInterface {
 public:
  Image(const Filename& filename)
    : score_(1), rotation_(-1), save_file_exists_(false) {
    filenames_.insert(filename);
    load();
  }

  void load_for(const Filename &filename) {
    quickmemsave = refptr<ImageInterface>(this);
    string dbfile = picdb_name(filename);
    fprintf(stderr, "Loading from %s\n", dbfile.c_str());
    FILE *f = fopen(dbfile.c_str(), "r");
    if (f) {
      save_file_exists_ = true;
      while (!feof(f)) {
	string key = readstring(f);
	if (getc(f) != '=') break;
	string value = readstring(f);
	if (getc(f) != '\n') break;
	fprintf(stderr, "key = %s  value = %s\n", key.c_str(), value.c_str());
	if (key == "score") {
	  score_ = resolve(score_, atoi(value.c_str()), 1);
	} else if (key == "rotation") {
	  rotation_ = resolve(rotation_, atoi(value.c_str()), 1);
	} else if (key == "alias") {
	  filenames_.insert(Filename(value));
	}
      }
      fclose(f);
    }
  }

  void load() {
    load_for(filename_f());
  }


  bool save() {
    fprintf(stderr, "Saving....\n");
    if (!save_file_exists_ &&
	filenames_.size() == 1 &&
	score_ == 1 &&
	(rotation_ == -1 || rotation_ == exif_rotation())) {
      // All default values and no previous save file, no need to save.
      fprintf(stderr, "No need to save\n");
    }
    set<Filename> filenames;
    get_filenames(&filenames);
    for (set<Filename>::const_iterator i = filenames.begin();
	 i != filenames.end();
	 ++i) {
      string dbname = picdb_name(*i);
      char ext[32];
      sprintf(ext, ".tmp%d", rand() % 1000000);
      string tmpname = dbname + ext;
      FILE *f = fopen(tmpname.c_str(), "w");
      if (!f) {
	fclose(f);
	fprintf(stderr, "Failed to open %s for writing.\n", tmpname.c_str());
	continue;
      }
      if (fprintf(f, "score=%d\nrotation=%d\n", score_, rotation()) <= 0) {
	fclose(f);
	unlink(tmpname.c_str());
	fprintf(stderr, "writing score.\n");
	return false;
      }
      for (set<Filename>::const_iterator j = filenames.begin();
	   j != filenames.end();
	   ++j) {
	if (fprintf(f,"alias=") <= 0) {
	  fclose(f);
	  unlink(tmpname.c_str());
	  fprintf(stderr, "Writing alias failed.\n");
	  return false;
	}
	if (!writestring(f, j->path())) {
	  fclose(f);
	  unlink(tmpname.c_str());
	  fprintf(stderr, "writing path failed.\n");
	  return false;
	}
	if (fprintf(f, "\n") <= 0) {
	  fclose(f);
	  unlink(tmpname.c_str());
	  fprintf(stderr, "writing newline failed.\n");
	  return false;
	}
      }
      if (fclose(f)) {
	fprintf(stderr, "fclose failed.\n");
	unlink(tmpname.c_str());
	return false;
      }
      if (rename(tmpname.c_str(), dbname.c_str())) {
	fprintf(stderr, "rename failed.\n");
	unlink(tmpname.c_str());
	return false;
      }
    }
    fprintf(stderr, "Save done.\n");
    return true;
  }

  bool file_exists(const string& str) const {
    struct stat s;
    return !stat(str.c_str(), &s) && S_ISREG(s.st_mode);
  }

  string filename() const {
    set<Filename> filenames;
    get_filenames(&filenames);
    for (set<Filename>::const_iterator i = filenames.begin();
	 i != filenames.end();
	 ++i) {
      string path = i->path();
      if (file_exists(path)) return path;
    }
    return "";
  }

  Filename filename_f() const {
    set<Filename> filenames;
    get_filenames(&filenames);
    for (set<Filename>::const_iterator i = filenames.begin();
	 i != filenames.end();
	 ++i) {
      string path = i->path();
      if (file_exists(path)) return *i;
    }
    return Filename("/");
  }

  void add_score(int s) {
    load();
    score_ += s;
    save();
  }

  void rotate(int r) {
    load();
    rotation_ += r + 360;
    rotation_ %= 360;
    save();
  }
  int exif_rotation() {
    ExifInterface *exif = GetExifData(filename());
    int exifrotation = exif->GetRotation();
    delete exif;
    switch (exifrotation) {
      case 1: return 0;
      case 6: return 90;
      case 3: return 180;
      case 8: return 270;
      case 2: return 0;
      case 5: return 90;
      case 4: return 180;
      case 7: return 270;
    }
  }

  pair<int, int> size() {
    return GetExifData(filename())->size();
  }

  int score() { return score_; }
  int rotation() {
    if (rotation_ == -1) rotation_ = exif_rotation();
    return rotation_;
  }

  bool IsEqual(const string &file) const {
    if (filenames_.find(Filename(file)) != filenames_.end()) {
      return true;
    }
    bool ret = false;
    FILE *af = fopen(file.c_str(), "rb");
    if (!af) return false;
    FILE *bf = fopen(filename().c_str(), "rb");
    if (!bf) {
      fclose(bf);
      return false;
    }
    while (1) {
      int ac = getc(af);
      int bc = getc(bf);
      if (ac != bc) break;
      if (ac == EOF) {
	ret = true;
	break;
      }
    }
    fclose(af);
    fclose(bf);
    return ret;
  }

  void AddFilename(const Filename &filename) {
    pthread_mutex_lock(&global_filename_lock);
    bool inserted = filenames_.insert(filename).second;
    pthread_mutex_unlock(&global_filename_lock);
    load_for(filename);
    if (inserted) save();
  }

  void get_filenames(std::set<Filename> *filenames) const {
    Locker locker(&global_filename_lock);
    *filenames = filenames_;
  }

  bool save_memory(string *str) {
    bool found = false;
    Locker locker(&global_filename_lock);
    for (set<Filename>::iterator i = filenames_.begin();
	 i != filenames_.end();
	 ++i) {
      if (i->dir == *str) {
	*str = i->dir;
	found = true;
	break;
      }
      if (i->file == *str) {
	*str = i->file;
	found = true;
	break;
      }
    }
    return found;
  }

  unsigned long long hash() const {
    std::set<Filename> filenames;
    get_filenames(&filenames);
    unsigned long long id = 0;
    for (set<Filename>::iterator i = filenames.begin();
	 i != filenames.end();
	 ++i) {
      id = max(id, i->hash());
    }
    return id;
  }

 private:
  int score_;
  int rotation_;
  set<Filename> filenames_;
  bool save_file_exists_;
};

class PicDB {
  typedef map<unsigned long long, refptr<ImageInterface> > ImageMap;

 public:
  ~PicDB() {
  }

  refptr<ImageInterface> GetRandomImage() const {
    int sum = 0;
    for (ImageMap::const_iterator i = images_.begin(); i != images_.end(); ++i) {
      sum += max(0, i->second->score());
    }
    if (sum == 0) return refptr<ImageInterface>(NULL);
    int n = rand() % sum;
    for (ImageMap::const_iterator i = images_.begin(); i != images_.end(); ++i) {
      if (n > i->second->score()) {
	n -= max(0, i->second->score());
      } else {
	return i->second;
      }
    }
    assert(false);
  }

  bool save_memory(string *str) {
    for (ImageMap::iterator i = images_.begin(); i != images_.end(); ++i) {
      if (i->second->save_memory(str)) return true;
    }
    return false;
  }

  void AddImage(ImageInterface *i) {
    unsigned long long h = i->hash();
    while (true) {
      ImageMap::iterator i = images_.find(h);
      if (i != images_.end()) {
	fprintf(stderr, " DUPLICATE IMAGE HASH, NEED STRONGER HASH FUNCTION!!!!\n");
	h++;
      } else {
	break;
      }
    }
    images_[h] = refptr<ImageInterface>(i);
    fprintf(stderr, "Images: %d\n", images_.size());
  }

  ImageStats GetImageStats() const {
    ImageStats ret;
    for (ImageMap::const_iterator i = images_.begin(); i != images_.end(); ++i) {
      ret.num_images++;
      int score = i->second->score();
      if (score <= 0) ret.num_downvotes++;
      if (score > 1) ret.num_upvotes += score - 1;
    }
    return ret;
  }

  refptr<ImageInterface> GetImageByHash(unsigned long long hash) {
    ImageMap::iterator i = images_.find(hash);
    if (i != images_.end()) {
      return i->second;
    } else {
      // Primary hash may have changed, do a _slow_ search.
      // This can happen if scanning reveals more copies of images, or if the user
      // saved a link to an image, and made more copies afterwards. We only index
      // primary hashes, but this let's us find it even if a secondary hash is used.
      for (ImageMap::const_iterator i = images_.begin(); i != images_.end(); ++i) {
	set<Filename> filenames;
	i->second->get_filenames(&filenames);
	for (set<Filename>::const_iterator j = filenames.begin();
	     j != filenames.end();
	     ++j) {
	  if (j->hash() == hash) {
	    // Found it!
	    refptr<ImageInterface> ret = i->second;
	    unsigned long long new_hash = ret->hash();
	    if (new_hash != i->first) {
	      // Put it in the correct place so we can find it next time.
	      images_.erase(i->first);
	      // Iterator now invalid!
	      images_[new_hash] = ret;
	    }
	    return ret;
	  }
	}
      }
      return refptr<ImageInterface>(NULL);
    }
  }

  refptr<ImageInterface> GetNextImage(refptr<ImageInterface> image) const {
    Filename filename = image->filename_f();
    Filename best_filename("/");
    refptr<ImageInterface> ret(NULL);
    for (ImageMap::const_iterator i = images_.begin(); i != images_.end(); ++i) {
      set<Filename> filenames;
      i->second->get_filenames(&filenames);
      for (set<Filename>::const_iterator j = filenames.begin();
	   j != filenames.end();
	   ++j) {
	if (*j <= filename) continue;
	if (!ret || *j < best_filename) {
	  best_filename = *j;
	  ret = i->second;
	}
      }
    }
    return ret;
  }

  refptr<ImageInterface> GetPrevImage(refptr<ImageInterface> image) const {
    Filename filename = image->filename_f();
    Filename best_filename("/");
    refptr<ImageInterface> ret(NULL);
    for (ImageMap::const_iterator i = images_.begin(); i != images_.end(); ++i) {
      set<Filename> filenames;
      i->second->get_filenames(&filenames);
      for (set<Filename>::const_iterator j = filenames.begin();
	   j != filenames.end();
	   ++j) {
	if (*j >= filename) continue;
	if (!ret || *j > best_filename) {
	  best_filename = *j;
	  ret = i->second;
	}
      }
    }
    return ret;
  }

  bool Compare(refptr<ImageInterface> &a, refptr<ImageInterface> &b) {
    // fprintf(stderr, "COMPARE: %d, %d,  %llu, %llu\n",
    // a->score(), b->score(),
    // a->hash(), b->hash());
    if (a->score() > b->score()) return false;
    if (a->score() < b->score()) return true;
    return a->hash() < b->hash();
  }

  void SWAP(refptr<ImageInterface> &a, refptr<ImageInterface> &b) { a.swap(b); }

  void PartialSort(vector<refptr<ImageInterface> > *images,
		   int range_start,
		   int range_end,
		   int first,
		   int last) {
    // fprintf(stderr, "PartialSort(%d, %d, %d, %d);\n", range_start, range_end, first, last);
    if (range_end <= first) return;
    if (last <= range_start) return;
    if (range_end - range_start < 2) return;
    if (range_end - range_start == 2) {
      if (Compare(images->at(range_start), images->at(range_start + 1))) {
	SWAP(images->at(range_start), images->at(range_start + 1));
      }
      return;
    }
    int pivot = (range_end + range_start) / 2;
    SWAP(images->at(range_start), images->at(pivot));
    pivot = range_start;
    int top = range_start + 1;
    int bottom = range_end - 1;
    // Normally I'd do some fancy stuff to make sure top and bottom move towards
    // the middle if all elemnts are equal, but since we know that it will be
    // extremely unusal for elements to be equal, why bother?

    while (top <= bottom) {
      while (top <= bottom && !Compare(images->at(top), images->at(pivot))) {
	top++;
      }
      while (top <= bottom && Compare(images->at(bottom), images->at(pivot))) {
	bottom--;
      }
      if (top < bottom) {
	SWAP(images->at(top), images->at(bottom));
	top++;
	bottom--;
      }
    }
    top--;
    if (top > pivot) {
      SWAP(images->at(pivot), images->at(top));
    }
    PartialSort(images, range_start, top, first, last);
    PartialSort(images, top + 1, range_end, first, last);
  }

  void GetTopList(int first, int last,
		  vector<refptr<ImageInterface> > *out) {
    vector<refptr<ImageInterface> > images(images_.size());
    size_t x = 0;
    for (ImageMap::const_iterator i = images_.begin(); i != images_.end(); ++i) {
      images[x++] = i->second;
    }
    PartialSort(&images, 0, images.size(), first, last);
    for (int i = max(0, first); i < last && i < images_.size(); i++) {
      out->push_back(images[i]);
    }
  }

  size_t num_images() const {
    ImageStats stats = GetImageStats();
    return stats.num_images - stats.num_downvotes;
  }

 private:
  ImageMap images_;
};

class Scanner {
public:
  Scanner(const vector<string>& to_scan) {
    picdb_ = new PicDB();
    for (int i = 0; i < to_scan.size(); i++) {
      AddToScan(to_scan[i]);
    }
  }
  typedef std::multimap<size_t, Image *> SizeMap;
  SizeMap size_to_files;

  void FoundImage(Filename f, size_t size) {
    pair<SizeMap::iterator, SizeMap::iterator> iters =
      size_to_files.equal_range(size);
    string path = f.path();
    size_t n = 0;
    for (SizeMap::iterator i = iters.first;
	 i != iters.second;
	 ++i) {
      n++;
    }
    fprintf(stderr, "    %d images of same size\n", n);
    for (SizeMap::iterator i = iters.first;
	 i != iters.second;
	 ++i) {
      if (i->second->IsEqual(path)) {
	fprintf(stderr, "Duplicate found\n");
	i->second->AddFilename(f);
	return;
      }
    }
    Image *i = new Image(f);
    size_to_files.insert(pair<size_t, Image *>(size, i));
    Locker locker(&db_lock);
    picdb_->AddImage(i);
  }

  bool endswith(string x, string y) {
    return x.rfind(y) == x.size() - y.size();
  }

  bool run() {
    if (to_scan_.empty()) return false;

    // Pick a random element from the scan queue.
    int index = rand() % to_scan_.size();
    Filename filename = to_scan_[index];
    to_scan_[index] = to_scan_[to_scan_.size() - 1];
    to_scan_.pop_back();

    string path = filename.path();
    fprintf(stderr, ":::%s\n", path.c_str());
    struct stat s;
    if (!stat(path.c_str(), &s)) {
      if (S_ISDIR(s.st_mode)) {
	DIR *dir = opendir(path.c_str());
	  
	while (struct dirent *entry = readdir(dir)) {
	  string filename(entry->d_name);
	  if (filename == ".") continue;
	  if (filename == "..") continue;
	  if (filename == ".xvpics") continue;
	  if (filename == ".thumbcache") continue;
	  if (filename == ".thumbnails") continue;
	  to_scan_.push_back(Filename(path, filename));
	}
	closedir(dir);
      }
      if (S_ISREG(s.st_mode)) {
	if (endswith(path, ".jpeg") ||
	    endswith(path, ".jpg") ||
	    endswith(path, ".JPG") ||
	    endswith(path, ".JPEG")) {
	  FoundImage(filename, s.st_size);
	}
      }
    }
  }

  void AddToScan(const string &file) {
    to_scan_.push_back(Filename(file));
  }
  PicDB *db() {
    return picdb_;
  }

  bool done() {
    return to_scan_.empty();
  }
private:
  PicDB *picdb_;
  vector<Filename> to_scan_;
};

Scanner *scanner = NULL;
PicDB *current_db = NULL;
vector<string> directories_to_scan;
time_t last_scan = 0;
int scan_timeout = 3600;

bool save_memory(string *str) {
  refptr<ImageInterface> tmp = quickmemsave;
  if (tmp && tmp->save_memory(str)) return true;
  if (current_db && current_db->save_memory(str)) return true;
  if (scanner && scanner->db() && scanner->db() != current_db &&
      scanner->db()->save_memory(str))
    return true;
  return false;
}

void InitScanner(const vector<string> &dirs,
		 const string &db_dir_,
		 int scan_timeout_) {
  directories_to_scan = dirs;
  db_dir = db_dir_;
  scan_timeout = scan_timeout_;
}

void StartScan() {
  if (scanner) return;
  scanner = new Scanner(directories_to_scan);
}

bool Scan() {
  if (scanner) {
    if (scanner->run()) {
      return true;
    }
  }
  return false;
}

// History tracking.
int current_image = 0;
int image_num = 0;
map<int, pair<refptr<ImageInterface>, time_t> > history;

refptr<ImageInterface> GetRandomImage() {
  fprintf(stderr, "GetRandomImage....\n");
  int required_images = 50;
  if (!current_db) {
    StartScan();
    current_db = scanner->db();
    required_images = 1;
  }
  fprintf(stderr, "Scan....\n");
  while (current_db->num_images() < required_images && scanner && !scanner->done()) {
    Scan();
  }
  if (current_db->num_images() == 0 && scanner->done()) {
    fprintf(stderr, "No images found!\n");
    exit(1);
  }
  if (scanner) {
    if (scanner->done()) {
      if (current_db != scanner->db()) {
	fprintf(stderr, "Deleting old DB\n");
	delete current_db;
      }
      fprintf(stderr, "Deleting scanner.\n");
      current_db = scanner->db();
      delete scanner;
      scanner = 0;
      last_scan = time(NULL);
    }
  } else {
    if (time(NULL) - last_scan > 3600) {
      StartScan();
    }
  }
  fprintf(stderr, "current_db_->GetRandomImage...\n");
  refptr<ImageInterface> ret = current_db->GetRandomImage();
  Locker locker(&history_lock);
  history[image_num] = pair<refptr<ImageInterface>, time_t>(ret, time(NULL));
  history.erase(image_num - 1000);
  image_num++;
  fprintf(stderr, "return random image...\n");
  return ret;
}

void SetCurrentImage(refptr<ImageInterface> image) {
  Locker locker(&history_lock);
  for (map<int, pair<refptr<ImageInterface>, time_t> >::reverse_iterator i = history.rbegin();
       i != history.rend();
       ++i) {
    if (i->second.first == image) {
      current_image = i->first;
    }
  }
}

int GetCurrentHistoryPos() {
  return current_image;
}

ImageStats GetImageStats() {
  Locker locker(&db_lock);
  if (!current_db) {
    ImageStats ret;
    return ret;
  }
  return current_db->GetImageStats();
}

refptr<ImageInterface> GetHistoryImage(int pos) {
  Locker locker(&history_lock);
  map<int, pair<refptr<ImageInterface>, time_t> >::iterator i = history.find(pos);
  if (i == history.end()) {
    return refptr<ImageInterface>(NULL);
  } else {
    return i->second.first;
  }
}

refptr<ImageInterface> GetImageByHash(unsigned long long hash) {
  Locker locker(&db_lock);
  if (!current_db) {
    return refptr<ImageInterface>(NULL);
  }
  return current_db->GetImageByHash(hash);
}

refptr<ImageInterface> GetNextImage(refptr<ImageInterface> image) {
  Locker locker(&db_lock);
  if (!current_db) {
    return refptr<ImageInterface>(NULL);
  }
  return current_db->GetNextImage(image);
}

refptr<ImageInterface> GetPrevImage(refptr<ImageInterface> image) {
  Locker locker(&db_lock);
  if (!current_db) {
    return refptr<ImageInterface>(NULL);
  }
  return current_db->GetPrevImage(image);
}

time_t GetHistoryTime(int pos) {
  Locker locker(&history_lock);
  map<int, pair<refptr<ImageInterface>, time_t> >::iterator i = history.find(pos);
  if (i == history.end()) {
    return 0;
  } else {
    return i->second.second;
  }
}

void GetTopList(int first, int last, vector<refptr<ImageInterface> > *out) {
  Locker locker(&db_lock);
  if (!current_db) return;
  return current_db->GetTopList(first, last, out);
}

