#ifndef PICDB_H
#define PICDB_H

#include <string>
#include <vector>
#include <set>
#include <utility>
#include "refptr.h"

bool save_memory(std::string *str);

struct Filename {
  explicit Filename(const std::string &file_) {
    size_t i = file_.rfind('/');
//    fprintf(stderr, " FOOO: %s\n", file_.c_str());
    if (file_ != "" && file_ != "/") {
      dir = file_.substr(0, i);
      file = file_.substr(i + 1);
      save_memory(&dir);
      save_memory(&file);
    }
  }
  Filename(const std::string &dir_, const std::string &file_)
  : dir(dir_), file(file_) {
    save_memory(&dir);
    save_memory(&file);
  }
  std::string path() const {
    return dir + "/" + file;
  }
  std::string dir;
  std::string file;

  bool operator==(const Filename &f) const { return file == f.file && dir == f.dir; }
  bool operator!=(const Filename &f) const { return file != f.file || dir != f.dir; }
  bool operator<(const Filename &f) const {
    return dir == f.dir ? file < f.file : dir < f.dir;
  }
  bool operator<=(const Filename &f) const {
    return dir == f.dir ? file <= f.file : dir < f.dir;
  }
  bool operator>(const Filename &f) const {
    return dir == f.dir ? file > f.file : dir > f.dir;
  }
  bool operator>=(const Filename &f) const {
    return dir == f.dir ? file >= f.file : dir > f.dir;
  }
  unsigned long long hash() const {
    unsigned long long id = 17;
    for (size_t i = 0; i < dir.size(); i++) id = id * 997 + dir[i];
    id *= 33;
    for (size_t i = 0; i < file.size(); i++) id = id * 997 + file[i];
    return id;
  }
};


class ImageInterface : public RefCounted {
public:
  ~ImageInterface() {}
  virtual std::string filename() const = 0;
  virtual Filename filename_f() const = 0;
  virtual int score() = 0;
  virtual int rotation() = 0;
  virtual std::pair<int, int> size() = 0;
  virtual void add_score(int s) = 0;
  virtual void rotate(int r) = 0;
  // Filename iterators
  virtual void get_filenames(std::set<Filename> *filenames) const = 0;
  virtual bool save_memory(std::string *str) = 0;
  virtual unsigned long long hash() const = 0;
};

// Each call to GetRandomImage() is 
refptr<ImageInterface> GetRandomImage();
void SetCurrentImage(refptr<ImageInterface> image);
int GetCurrentHistoryPos();
refptr<ImageInterface> GetHistoryImage(int pos);
refptr<ImageInterface> GetImageByHash(unsigned long long hash);
refptr<ImageInterface> GetNextImage(refptr<ImageInterface> image);
refptr<ImageInterface> GetPrevImage(refptr<ImageInterface> image);
time_t GetHistoryTime(int pos);

struct ImageStats {
  ImageStats(): num_images(0), num_upvotes(0), num_downvotes(0) {}
  size_t num_images;
  size_t num_upvotes;
  size_t num_downvotes;
};

ImageStats GetImageStats();
void GetTopList(int first, int last, std::vector<refptr<ImageInterface> > *out);

// Call repeatedly, returns false if no scanning is currently required.
bool Scan();

// Call from main() to specify where to scan.
void InitScanner(const std::vector<std::string> &dirs,
		 const std::string &db_dir,
		 int scan_timeout);

#endif
