#include <stdio.h>
#include <stdlib.h>
#include <setjmp.h>
#include <jpeglib.h>
#include <jerror.h>
#include <memory.h>

#include "superscaler.h"
#include "tuple.h"
#include "exif.h"
#include "picdb.h"

using namespace SuperScaler;
using namespace std;

struct my_error_mgr {
  struct jpeg_error_mgr pub;	/* "public" fields */

  jmp_buf setjmp_buffer;	/* for return to caller */
};

typedef struct my_error_mgr * my_error_ptr;


METHODDEF(void)
my_error_exit (j_common_ptr cinfo)
{
  /* cinfo->err really points to a my_error_mgr struct, so coerce pointer */
  my_error_ptr myerr = (my_error_ptr) cinfo->err;

  /* Always display the message. */
  /* We could postpone this until after returning, if we chose. */
  (*cinfo->err->output_message) (cinfo);

  /* Return control to the setjmp point */
  longjmp(myerr->setjmp_buffer, 1);
}

typedef Tuple<unsigned char, 3> Pixel;
typedef Tuple<int, 3> TmpPixel;

Matrix<float> create_scaler(int from, int to) {
  Filter *f = SuperScaler::LanczosFilter(3);
  Matrix<float> ret;
#if 1
  ret = SuperScaler::create_scaler(from, 0.0, from, to, f);
#else
  if (from <= to) {
    ret = SuperScaler::create_scaler(from, 0.0, from, to, f);
  } else {
    ret = SuperScaler::create_scaler(to, 0.0, to, from, f);
    ret = ret.left_invert();
  }
#endif
  delete f;
  return ret;
}

class RowStream {
public:
  virtual ~RowStream() {}
  virtual int rows() = 0;
  virtual int columns() = 0;
  virtual void GetRow(TmpPixel *dest) = 0;
  virtual bool failure() = 0;
};

class JpegRowStream : public RowStream {
public:
  explicit JpegRowStream(FILE *infile) : buffer(NULL), failure_(false) {
    /* We set up the normal JPEG error routines, then override error_exit. */
    cinfo.err = jpeg_std_error(&jerr.pub);
    jerr.pub.error_exit = my_error_exit;

    /* Establish the setjmp return context for my_error_exit to use. */
    if (setjmp(jerr.setjmp_buffer)) {
      return;
    }
    /* Now we can initialize the JPEG decompression object. */
    jpeg_create_decompress(&cinfo);
    
    jpeg_stdio_src(&cinfo, infile);
    (void) jpeg_read_header(&cinfo, TRUE);
    (void) jpeg_start_decompress(&cinfo);
    
    buffer = (*cinfo.mem->alloc_sarray)
      ((j_common_ptr) &cinfo, JPOOL_IMAGE, cinfo.output_width * cinfo.output_components, 1);
  }

  ~JpegRowStream() {
    /* Establish the setjmp return context for my_error_exit to use. */
    if (setjmp(jerr.setjmp_buffer)) {
      jpeg_destroy_decompress(&cinfo);
      return;
    }

    while (cinfo.output_scanline + 1 < cinfo.output_height && !failure_) {
      (void) jpeg_read_scanlines(&cinfo, buffer, 1);
    }

    (void) jpeg_finish_decompress(&cinfo);
    jpeg_destroy_decompress(&cinfo);
  }

  struct jpeg_decompress_struct cinfo;
  struct my_error_mgr jerr;
  JSAMPARRAY buffer;		/* Output row buffer */
  bool failure_;

  virtual int rows() { return cinfo.output_height; }
  virtual int columns() { return cinfo.output_width; }
  virtual bool failure() { return failure_; }

  virtual void GetRow(TmpPixel *dest) {
    if (failure_) return;
    if (cinfo.output_scanline >= cinfo.output_height)
      return;

    /* Establish the setjmp return context for my_error_exit to use. */
    if (setjmp(jerr.setjmp_buffer)) {
      failure_ = true;
      return;
    }

    std::vector<TmpPixel> row(cinfo.output_width);
    (void) jpeg_read_scanlines(&cinfo, buffer, 1);
    for(int x = 0; x < columns(); x++) {
      if (cinfo.output_components == 3) {
	dest[x] = TmpPixel(buffer[0][x * 3], buffer[0][x * 3 + 1], buffer[0][x * 3 + 2]);
      }
      if (cinfo.output_components == 1) {
	dest[x] = TmpPixel(buffer[0][x], buffer[0][x], buffer[0][x]);
      }
    }
  }
};

class CropRowStream : public RowStream {
public:
  CropRowStream(RowStream *row_stream,
		int top,
		int bottom,
		int left,
		int right)
    : row_stream_(row_stream),
      top_(top),
      bottom_(bottom),
      left_(left),
      right_(right),
      current_(0) {
    TmpPixel tmp[row_stream_->columns()];
    for (int i = 0; i < top_; i++) {
      row_stream_->GetRow(tmp);
    }
  }
  ~CropRowStream() {
    TmpPixel tmp[row_stream_->columns()];
    for (int i = 0; i < bottom_; i++) {
      row_stream_->GetRow(tmp);
    }
    delete row_stream_;
  }
  int rows() {
    return row_stream_->rows() - top_ - bottom_;
  }
  int columns() {
    return row_stream_->columns() - left_ - right_;
  }

  virtual void GetRow(TmpPixel *dest) {
    assert(current_ <= rows());
    if (left_ == 0 && right_ == 0) {
      row_stream_->GetRow(dest);
    } else {
      TmpPixel tmp[row_stream_->columns()];
      row_stream_->GetRow(tmp);
      memcpy(dest, tmp + left_,
	     sizeof(TmpPixel) * columns());
    }
    current_++;
  }

  virtual bool failure() { return row_stream_->failure(); }

  RowStream *row_stream_;
  int top_;
  int bottom_;
  int left_;
  int right_;
  int current_;
};

class ScaleRowStream : public RowStream {
public:
  ScaleRowStream(RowStream *row_stream,
		 int width,
		 int height)
    : row_stream_(row_stream),
      width_(width),
      height_(height),
      buffer_rows_(0),
      input_row_(0),
      output_row_(0),
      output_buffer_bottom_(0) {
    row_scaler_.copyfrom(create_scaler(row_stream->columns(), width) * 2048.0);
    col_scaler_.copyfrom(create_scaler(row_stream->rows(), height) * 2048.0);
    for (int y = 0; y < col_scaler_.rows(); y++) {
      buffer_rows_ = std::max(buffer_rows_, col_scaler_.row(y).width() + 1);
    }
  }
  ~ScaleRowStream() {
    TmpPixel tmp[row_stream_->columns()];
    while (input_row_ < row_stream_->rows()) {
      row_stream_->GetRow(tmp);
      input_row_++;
    }
    delete row_stream_;
  }
  int rows() {
    return height_;
  }
  int columns() {
    return width_;
  }

  virtual void GetRow(TmpPixel *dest) {
    if (output_row_ >= rows()) return;
    std::vector<TmpPixel> row(row_stream_->columns());
    while (col_scaler_.row(output_row_).largest_used_column() >= input_row_) {
      row_stream_->GetRow(&row[0]);
//      fprintf(stderr, " INROW: %d ", input_row_);
      row_buffer_[input_row_++] = row_scaler_ * row;
      if (row_buffer_.size() > buffer_rows_) {
//	fprintf(stderr, "<ERASE %d>", output_buffer_bottom_);
	row_buffer_.erase(output_buffer_bottom_++);
      }
    }

//    fprintf(stderr, "\nOUTPUT ROW: %d\n", output_row_);
    Matrix<TmpPixel> out(1, row_scaler_.rows()); 
    for (Row<int>::const_iterator i = col_scaler_.row(output_row_).begin();
	 i != col_scaler_.row(output_row_).end();
	 ++i) {
//      fprintf(stderr, "(%d)", i->first);
    }
    for (Row<int>::const_iterator i = col_scaler_.row(output_row_).begin();
	 i != col_scaler_.row(output_row_).end();
	 ++i) {
      assert(row_buffer_.find(i->first) != row_buffer_.end());
      out.addmultiplied(row_buffer_[i->first], i->second);
    }
    for (int x = 0; x < row_scaler_.rows(); x++) {
      dest[x][0] = clamp(out.get(0, x)[0] >> 22, 0, 255);
      dest[x][1] = clamp(out.get(0, x)[1] >> 22, 0, 255);
      dest[x][2] = clamp(out.get(0, x)[2] >> 22, 0, 255);
    }
    output_row_++;
  }

  virtual bool failure() { return row_stream_->failure(); }

  map<int, Matrix<TmpPixel> > row_buffer_;
  RowStream *row_stream_;
  Matrix<int> row_scaler_;
  Matrix<int> col_scaler_;
  int width_;
  int height_;
  int buffer_rows_;
  int input_row_;
  int output_row_;
  int output_buffer_bottom_;
};

class BoxRowStream : public RowStream {
public:
  BoxRowStream(RowStream *row_stream,
		int top,
		int bottom,
		int left,
		int right)
    : row_stream_(row_stream),
      top_(top),
      bottom_(bottom),
      left_(left),
      right_(right),
      current_(0) {
  }
  ~BoxRowStream() {
    delete row_stream_;
  }
  int rows() {
    return row_stream_->rows() + top_ + bottom_;
  }
  int columns() {
    return row_stream_->columns() + left_ + right_;
  }

  virtual void GetRow(TmpPixel *dest) {
    if (current_ < top_ || current_ >= rows() - bottom_) {
      for (int i = 0; i < columns(); i++) dest[i] = TmpPixel(0);
    } else {
      for (int i = 0; i < left_; i++) dest[i] = TmpPixel(0);
      for (int i = columns() - right_; i < columns(); i++) dest[i] = TmpPixel(0);
      row_stream_->GetRow(dest + left_);
    }
    current_++;
  }

  virtual bool failure() { return row_stream_->failure(); }

  RowStream *row_stream_;
  int top_;
  int bottom_;
  int left_;
  int right_;
  int current_;
};

RowStream *ClipAndScale(RowStream *stream, int target_x, int target_y, int max_crop) {
  fprintf(stderr, "Target: %d x %d\n", target_x, target_y);
  int x = stream->columns();
  int y = stream->rows();
  fprintf(stderr, "Input: %d x %d\n", x, y);
  if (x == target_x && y == target_y) return stream;
  double x_scale = target_x / (double)x;
  double y_scale = target_y / (double)y;
  fprintf(stderr, "DIFF: %f\n", x_scale / y_scale);
  fprintf(stderr, "DIFF: %f\n", y_scale / x_scale);
  double scale;
  // Which is better???
  // scale = std::min(x_scale, y_scale);
  // fprintf(stderr, "scale = %f\n", scale);
  // scale = 2/(1/x_scale + 1/y_scale);
  // fprintf(stderr, "scale = %f\n", scale);
  // scale = sqrt(x_scale * y_scale);
  // fprintf(stderr, "scale = %f\n", scale);
  // scale = (x_scale + y_scale) / 2;
  // fprintf(stderr, "scale = %f\n", scale);
  // scale = std::max(x_scale, y_scale);
  // fprintf(stderr, "scale = %f\n", scale);

  double wiggle = (100 + max_crop) / 100.0;
  scale = std::max(std::min(x_scale * wiggle, y_scale),
		   std::min(y_scale * wiggle, x_scale));

  fprintf(stderr, "x_scale: %f  y_scale: %f   final_scale: %f\n", 1/x_scale, 1/y_scale, 1/scale);
  int x_crop = std::max(0, x - (int)(target_x / scale));
  int y_crop = std::max(0, y - (int)(target_y / scale));
  fprintf(stderr, "x crop: %d   y crop: %d\n", x_crop, y_crop);
  if (x_crop || y_crop) {
    stream = new CropRowStream(stream,
			       y_crop / 2, y_crop - y_crop / 2,
			       x_crop / 2, x_crop - x_crop / 2);
    x -= x_crop;
    y -= y_crop;
    fprintf(stderr, "CROPPED: %d x %d\n", stream->columns(), stream->rows());
  }
  fprintf(stderr, "SCALE: %f x %f\n", x * scale, y * scale);
  int scaled_x = std::min(target_x, (int)(x * scale + 0.5));
  int scaled_y = std::min(target_y, (int)(y * scale + 0.5));
  stream = new ScaleRowStream(stream, scaled_x, scaled_y);
  fprintf(stderr, "SCALED: %d x %d\n", stream->columns(), stream->rows());
  assert(target_x >= stream->columns());
  assert(target_y >= stream->rows());
  int x_box = target_x - scaled_x;
  int y_box = target_y - scaled_y;
  fprintf(stderr, "x_box: %d y_box: %d\n", x_box, y_box);
  stream = new BoxRowStream(stream,
			    y_box / 2, y_box - y_box / 2,
			    x_box / 2, x_box - x_box / 2);
  
  fprintf(stderr, "BOXED: %d x %d\n", stream->columns(), stream->rows());
  assert(target_x == stream->columns());
  assert(target_y == stream->rows());
  return stream;
}

int get_rotation_for_file(const string &filename) {
  ExifInterface *exif = GetExifData(filename);
  int ret = exif->GetRotation();
  delete exif;
  fprintf(stderr, "ROTATION=%d\n", ret);
  return ret;
}


void stream_to_p6(RowStream *stream) {
  printf("P6\n%d %d\n255\n", stream->columns(), stream->rows());
  TmpPixel row[stream->columns()];
  for (int y = 0; y < stream->rows(); y++) {
    stream->GetRow(row);
    for (int x = 0; x < stream->columns(); x++) {
      printf("%c%c%c", row[x][0], row[x][1], row[x][2]);
    }
  }
  delete stream;
}

class ExifOrientation {
public:
  explicit ExifOrientation(int o) : orientation_(o) {
  }
  ExifOrientation rotate90clockwise() const {
    switch (orientation_) {
      case 1: return ExifOrientation(8);
      case 8: return ExifOrientation(3);
      case 3: return ExifOrientation(6);
      case 6: return ExifOrientation(1);
      case 2: return ExifOrientation(7);
      case 7: return ExifOrientation(4);
      case 4: return ExifOrientation(5);
      case 5: return ExifOrientation(2);
    }
  }
  ExifOrientation mirror() const {
    return ExifOrientation(orientation_ ^ 1);
  }
private:
  int orientation_;
};

// Load, crop & rotate jpeg.
bool load_jpeg(refptr<ImageInterface> image,
	       int xsize,
	       int ysize,
	       int max_crop,
	       unsigned char *data) {
  string filename = image->filename();
  fprintf(stderr, "loading %s\n", filename.c_str());
  int row_jump;
  int col_jump;
  switch (image->rotation()) {
    default: row_jump = xsize; col_jump = 1; break;
    case 90: row_jump = -1; col_jump = xsize; break; 
    case 180: row_jump = -xsize; col_jump = -1 ; break;
    case 270: row_jump = 1; col_jump = -xsize; break;
  }
#if 0
  switch (get_rotation_for_file(filename)) {
    case 1: row_jump = xsize; col_jump = 1; break;
    case 2: row_jump = xsize; col_jump = -1; break;
    case 3: row_jump = -xsize; col_jump = -1 ; break;
    case 4: row_jump = -xsize; col_jump = 1; break;
    case 5: row_jump = 1; col_jump = xsize; break;
    case 6: row_jump = -1; col_jump = xsize; break;  // 90 degrees
    case 7: row_jump = -1; col_jump = -xsize; break;
    case 8: row_jump = 1; col_jump = -xsize; break;
  }
#endif
  row_jump *= 4;
  col_jump *= 4;

  FILE *f = fopen(filename.c_str(), "rb");
  RowStream *stream = new JpegRowStream(f);
  if (row_jump >= -4 && row_jump <= 4) {
    stream = ClipAndScale(stream, ysize, xsize, max_crop);
  } else {
    stream = ClipAndScale(stream, xsize, ysize, max_crop);
  }

  unsigned char *row_base = data;
  if (row_jump < 0) {
    row_base -= (stream->rows() - 1) * row_jump;
  }
  if (col_jump < 0) {
    row_base -= (stream->columns() - 1) * col_jump;
  }

  bool ret = true;
  TmpPixel row[stream->columns()];
  for (int y = 0; y < stream->rows(); y++) {
    fprintf(stderr, "\r %d / %d ", y, stream->rows());
    unsigned char *pos = row_base;
    stream->GetRow(row);
    if (stream->failure()) {
      ret = false;
      break;
    }
    for (int x = 0; x < stream->columns(); x++) {
      assert(pos >= data);
      assert(pos + 3 < data + xsize * ysize * 4);
      pos[0] = row[x][0];
      pos[1] = row[x][1];
      pos[2] = row[x][2];
      pos[3] = 255;
      pos += col_jump;
    }
    row_base += row_jump;
  }
  fprintf(stderr, "\n");
  delete stream;
  fclose(f);
  return ret;
}

