#ifndef SUPERSCALER_MATRIX_H
#define SUPERSCALER_MATRIX_H

#include <map>
#include <vector>
#include <math.h>
#include <assert.h>
#include <stdio.h>
#include <iostream>
#include "tuple.h"

namespace SuperScaler {

template<class T> static inline bool very_close(T a, T b) { return a == b; }
#ifndef double
template<> inline bool very_close(double a, double b) { return fabs(a-b) < 0.0000000000001; }
#endif
template<> inline bool very_close(float a, float b) { return fabs(a-b) < 0.00000001; }

template<class T> T clamp(T v, T low, T high) {
  if(v < low) return low;
  if(v > high) return high;
  return v;
}
#if 0
template<class T> T clamp(T v, int low, int high) {
  if(v < low) return low;
  if(v > high) return high;
  return v;
}
#endif

template<class T> T abs(T v) {
  if (v < 0) return -v;
  return v;
}

template<class T> class Row {
public:
  typedef typename std::map<int, T>::iterator iterator;
  typedef typename std::map<int, T>::const_iterator const_iterator;
  const_iterator begin() const { return data_.begin(); }
  const_iterator end() const { return data_.end(); }
  iterator begin() { return data_.begin(); }
  iterator end() { return data_.end(); }

  T get(int x) const {
    const_iterator i = data_.find(x);
    if (i == data_.end()) return 0.0;
    return i->second;
  }

  void set(int x, T val) {
    if (very_close<T>(val, 0)) {
      data_.erase(x);
    } else {
      data_[x] = val;
    }
  }

  void add(int x, T val) {
    set(x, get(x) + val);
  }

  void operator*=(T v) {
    for (iterator i = data_.begin(); i != data_.end(); ++i) {
      i->second *= v;
    }
  }

  bool operator==(const Row& other) const {
    if (data_.size() != other.data_.size()) {
      return false;
    }
    const_iterator a = begin();
    const_iterator b = other.begin();
    while (a != end()) {
      if (b == other.end()) return false;
      if (a->first != b->first) return false;
      if (!very_close(a->second, b->second)) return false;
      ++a;
      ++b;
    }
    return b == other.end();
  }

  void muladd(const Row& other, T v) {
    for (const_iterator i = other.begin(); i != other.end(); ++i) {
      set(i->first, get(i->first) + i->second * v);
    }
  }

  void operator+=(const Row& other) {
    for (const_iterator i = other.begin(); i != other.end(); ++i) {
      set(i->first, get(i->first) + i->second);
    }
  }

  void zero() {
    typename std::map<int, T> tmp;
    tmp.swap(data_);
  }

  void swap(Row &other) {
    data_.swap(other.data_);
  }

  size_t memsize() const {
    return data_.size() * sizeof(typename std::pair<int, T>);
  }

  int largest_used_column() const {
    if (data_.empty()) return 0;
    return data_.rbegin()->first;
  }

  int smallest_used_column() const {
    if (data_.empty()) return 0;
    return data_.begin()->first;
  }

  int width() const {
    if (data_.empty()) return 0;
    return data_.rbegin()->first - data_.begin()->first;
  }

  void normalize() {
    T sum = 0;
    for (iterator i = data_.begin(); i != data_.end(); ++i) {
      sum += i->second;
    }
    sum = 1 / sum;
    for (iterator i = data_.begin(); i != data_.end(); ++i) {
      i->second *= sum;
    }
  }

private:
  std::map<int, T> data_;
};


template<class T> class Matrix {
 private:
  // We keep a circular list of matrices which
  // share the same data.
  Matrix *next_, *prev_;

 public:
  typedef Row<T> RowType;
  int columns_;

 private:
  std::vector<Row<T> > *data_;

 public:
  const Row<T> &row(int x) const {
    return data_->at(x);
  }

  Row<T> &row(int x) {
    return data_->at(x);
  }

  int rows() const {
    if (!data_) return 0;
    return data_->size();
  }

  int columns() const {
    return columns_;
  }

  Matrix() : 
    next_(this),
    prev_(this),
    columns_(0),
    data_(NULL) {
  }

  // Zero matrix
  Matrix(int xsize, int ysize) :
      next_(this),
      prev_(this),
      columns_(xsize),
      data_(new std::vector<Row<T> >(ysize)) {
    if (rows()) assert(columns_ > 0);
  }

  // square unity matrix
  explicit Matrix(int size) : 
      next_(this),
      prev_(this),
      columns_(size),
      data_(new std::vector<Row<T> >(size)) {
    if (rows()) assert(columns_ > 0);
    unity(1);
  }

  // Copy 
  Matrix(const Matrix &x) :
    next_(x.next_),
    prev_(x.next_->prev_),
    columns_(x.columns_),
    data_(x.next_->prev_->data_) {
    next_->prev_ = prev_->next_ = this;
    if (rows()) assert(columns_ > 0);
  }

  void free() {
    assert(next_->prev_ == this);
    assert(prev_->next_ == this);
    if(next_ == this) {
      delete data_;
    } else {
      // Unlink ourselves
      next_->prev_ = prev_;
      prev_->next_ = next_;
    }
  }

  // assignment
  Matrix& operator=(const Matrix &x) {
    if (&x != this) {
      free();
      next_ = x.next_;
      prev_ = x.prev_->prev_;
      next_->prev_ = prev_->next_ = this;
      columns_ = x.columns_;
      data_ = x.data_;
      if (rows()) assert(columns_ > 0);
    }
    return *this;
  }

  template<class X>
  Matrix& copyfrom(const Matrix<X> &x) {
    free();
    columns_ = x.columns();
    data_ = new std::vector<Row<T> >(x.rows());
    for (int y = 0; y < rows(); y++) {
      for (typename Row<X>::const_iterator i = x.row(y).begin();
	   i != x.row(y).end();
	   ++i) {
	set(i->first, y, i->second);
      }
    }
  }

  // Set diagonal to 'x'
  Matrix& operator=(const T &x) {
    unity(x);
  }

  ~Matrix() { free(); }

  void zero() {
    if (!data_) return;
    for (int e=0; e < data_->size(); e++) row(e).zero();
  }

  void unity(T val) {
    if (!data_) return;
    assert(data_->size() == columns_);
    zero();
    for(int e=0; e < columns_; e++) set(e, e, val); 
  }

  const T get(int x, int y) const {
    assert(x >= 0);
    assert(x < columns_);
    return row(y).get(x);
  }

  void set(int x, int y, T v) {
    assert(x >= 0);
    assert(x < columns_);
    return row(y).set(x, v);
  }

  void add(int x, int y, T v) {
    assert(x >= 0);
    assert(x < columns_);
    row(y).add(x, v);
  }

  void addclamped(int x, int y, T v) {
    row(clamp(y, 0, rows() - 1)).add(clamp(x, 0, columns() - 1), v);
  }

  const T clampedvalue(int x, int y) const {
    return get(clamp(x, 0, columns() - 1), clamp(y, 0, rows() - 1));
  }

  size_t memsize() const {
    size_t ret = 0;
    for (int y = 0; y < rows(); y++) {
      ret += row(y).memsize();
    }
    return ret;
  }

  bool operator==(const Matrix &m) const {
    if(m.data_ == data_) return true;
    if(rows() != m.rows()) return false;
    if(columns() != m.columns()) return false;

    for (int y = 0; y < rows(); y++) {
      if (!(row(y) == m.row(y)))
	return false;
    }

    return true;
  }

  Matrix copy() const {
    Matrix ret(rows(), columns());
    for (int y = 0; y < rows(); y++)
      for (typename Row<T>::const_iterator x = row(y).begin(); x != row(y).end(); ++x)
	ret.set(x->first, y, x->second);
    return ret;
  }

  Matrix operator*(T v) const {
    Matrix ret(columns(), rows());
    for (int y = 0; y < rows(); y++)
      for (typename Row<T>::const_iterator x = row(y).begin(); x != row(y).end(); ++x)
	ret.set(x->first, y, x->second * v);
    return ret;
  }

  template<class X>
  Matrix<X> operator*(const std::vector<X> &v) const {
    assert(columns() == v.size());
    Matrix<X> ret(1, rows());
    for(int y = 0; y < rows(); y++) {
      X tmp(0);
      for (typename Row<T>::const_iterator z = row(y).begin(); z != row(y).end(); ++z) {
	tmp += v[z->first] * z->second;
      }
      ret.set(0, y, tmp);
    }
    return ret;
  }

  template<class X>
  Matrix<X> operator*(const Matrix<X> &m) const {
    assert(columns() == m.rows());
    if (m.columns() == 1) {
      std::vector<X> lookup(m.rows());
      for (int i = 0; i < m.rows(); i++) lookup[i] = m.get(0, i);
      return (*this) * lookup;
    } else {
      Matrix<X> ret(m.columns(), rows());
      for(int y = 0; y < rows(); y++) {
	for (typename Row<T>::const_iterator z = row(y).begin(); z != row(y).end(); ++z) {
	  for (typename Row<X>::const_iterator x = m.row(z->first).begin(); x != m.row(z->first).end(); ++x) {
	    ret.add(x->first, y, x->second * z->second);
	  }
	}
      }
      return ret;
    }
  }
  
  void operator+=(const Matrix& other) {
    assert(columns() == other.columns());
    assert(rows() == other.rows());
    for (int y = 0; y < rows(); y++)
      data_->at(y) += other.row(y);
  }

  void addmultiplied(const Matrix &other, T v) {
    assert(columns() == other.columns());
    assert(rows() == other.rows());
    for (int y = 0; y < rows(); y++)
      data_->at(y).muladd(other.row(y), v);
  }

  void normalizerows() {
    for (int y = 0; y < rows(); y++) row(y).normalize();
  }

  void swaprows(int a, int b) {
    row(a).swap(row(b));
  }

  void rowmuladd(int a, int b, T val) {
    row(a).muladd(row(b), val);
  }


  template<class X>
  bool solve(Matrix<X> &answers) {
    assert(rows() == columns());
    assert(rows() == answers.rows());
    for(int x=0; x < columns(); x++) {
      // fprintf(stderr, "COL: %d\n", x);
      // PrintEq(answers);
      int best_row = -1;
      T best_val = 0;
      int best_row_length = columns();
      for (int y = x; y < rows(); y++) {
#if 0
	if (abs(get(x, y)) > 0.0001 && row(y).largest_used_column() < best_row_length) {
	  best_val = get(x, y);
	  best_row = y;
	  best_row_length = row(y).largest_used_column();
	}
#endif
#if 1
	if (abs(get(x, y)) > abs(best_val)) {
	  best_val = get(x, y);
	  best_row = y;
	}
#endif
      }
      assert(best_row >= 0);
      if (best_row != x) {
	answers.swaprows(x, best_row);
	swaprows(x, best_row);
      }
      if (best_val != 1.0) {
	row(x) *= 1.0 / best_val;
	answers.row(x) *= 1.0 / best_val;
      }
      // PrintEq(answers);
      for(int y = x + 1; y < rows(); y++) {
	if (x == y) continue;
	if (get(x, y) == 0.0) continue;
	T tmp = - get(x, y);
	answers.rowmuladd(y, x, tmp);
	rowmuladd(y, x, tmp);
      }
    }
#if 1
    for(int x = columns() - 1; x >= 0; x--) {
      for(int y = x - 1; y >= 0; y--) {
	if (get(x, y) == 0.0) continue;
	T tmp = - get(x, y);
	answers.rowmuladd(y, x, tmp);
	rowmuladd(y, x, tmp);
      }
    }
#endif

    return true;
  }

  void Print() const {
    for (int y = 0; y < rows(); y++) {
      for (int x = 0; x < columns(); x++) {
	printf("%+7.2f ", get(x, y));
      }
      printf("\n");
    }
    printf("------------------------------\n");
  }

  void PrintEq(const Matrix &other) const {
    for (int y = 0; y < rows(); y++) {
      for (int x = 0; x < columns(); x++) {
	printf("%+7.2f ", get(x, y));
      }
      printf(" | ");
      for (int x = 0; x < other.columns(); x++) {
	printf("%+7.2f ", other.get(x, y));
      }
      printf("\n");
    }
    printf("------------------------------\n");
  }

  Matrix solve_least_square(const Matrix<T> &answers) const {
    assert(answers.rows() == rows());
    fprintf(stderr, "transpose...\n");
    Matrix tmp1 = this->transpose();
    fprintf(stderr, "mult1...\n");
    Matrix tmp2 = tmp1 * *this;
    fprintf(stderr, "mult2...\n");
    Matrix tmp3 = tmp1 * answers;
    fprintf(stderr, "solve...\n");
    tmp2.solve(tmp3);
    fprintf(stderr, "solved...\n");
    return tmp3;
  }

  // FIXME
  Matrix transpose() const {
    Matrix ret(rows(), columns());
    for (int y = 0; y < rows(); y++) {
      for (typename Row<T>::const_iterator x = row(y).begin(); x != row(y).end(); ++x) {
	ret.set(y, x->first, x->second);
      }
    }

    return ret;
  }

  Matrix invert() const {
    Matrix ret(rows());
    copy().solve(ret);
    return ret;
  }

  Matrix left_invert() const {
#if 0
    Matrix t = transpose();
    Matrix aat = (*this) * t;
    return t * aat.invert();
#endif
#if 0
    return solve_least_square(Matrix(rows()));
#endif
#if 0
    fprintf(stderr, "transpose...\n");
    Matrix tmp1 = this->transpose();
    fprintf(stderr, "mult1...\n");
    Matrix tmp2 = tmp1 * *this;
    fprintf(stderr, "solve...\n");
//    tmp2.solve(tmp1);
    Matrix answers(tmp2.rows());
    tmp2.solve(answers);
    fprintf(stderr, "solved...\n");
    return answers * tmp1;
#endif
#if 1
    fprintf(stderr, "transpose...\n");
    Matrix tmp1 = this->transpose();
    fprintf(stderr, "mult1...\n");
    Matrix tmp2 = tmp1 * *this;
    fprintf(stderr, "solve...\n");
    tmp2.solve(tmp1);
    return tmp1;
#endif
  }
};
};

template<class T> std::ostream& operator<<(std::ostream& os,
					   const SuperScaler::Matrix<T> m) {
  os<<std::endl;
  for(int y = 0; y < m.rows();y++) {
    os << (y == 0 ? "/ " : y == m.rows() - 1 ? "\\ " : "| ");  
    for(int x = 0; x<m.columns();x++) {
      if(x != 0) os << ", ";
      os << m.get(x,y);
    }
    os << (y == 0 ? " \\\n" : y == m.columns() - 1 ? " /\n" : " |\n");  
  }
  return os;
}

#endif
