#include <stdio.h>
#include "sparse_matrix.h"
#include "superscaler.h"

namespace SuperScaler {

  class HT {
  public:
    Matrix<float> halftone;
    HT(int size) : halftone(size, size) {
      int num = 1;
      mkhalftonematrix(0, 0, 1, num);
    }

    void mkhalftonematrix(int x, int y, int size, int &num) {
      if(size > halftone.columns() && size > halftone.rows()) {
	if(x < halftone.columns() && y < halftone.rows())
	  halftone.set(x, y, num++ /
		       (float)halftone.columns() / halftone.rows());
      } else {
	mkhalftonematrix(x,y,size*2,num);
	mkhalftonematrix(x+size,y+size,size*2,num);
	mkhalftonematrix(x+size,y,size*2,num);
	mkhalftonematrix(x,y+size,size*2,num);
      }
    }
  };

  inline double sinc(double f) {
    if(f==0.0) return 1.0;
    f*=M_PI;
    // if(f==0.0) return 1.0;
    f=sin(f)/f;
    // if(!isnormal(f)) return 1.0;
    return f;
  }
  
  class SincFilter : public Filter
  {
  public:
    virtual double filter_size(double rate) {
      return (1<<16) * (rate > 1.0 ? rate : 1.0);;
    }
    
    virtual double filter(double dist) {
      return sinc(dist);
    }
  };

  class SphericalBessel1Filter : public Filter
  {
  public:
    virtual double filtersize(double rate) {
      return (1<<16) * (rate > 1.0 ? rate : 1.0);;
    }
    
    virtual double filter(double dist) {
      if(dist == 0.0) return 0.0; 
      double x=dist * M_PI;
      return (sinc(dist) - cos(x))/x;
    }
  };
  
  class SphericalBessel2Filter : public Filter
  {
  public:
    virtual double filtersize(double rate) {
      return (1<<16) * (rate > 1.0 ? rate : 1.0);;
    }
    
    virtual double filter(double dist) {
      if(dist == 0.0) return 0.0; 
      double x=dist * M_PI;
      return (3/(x*x*x) - 1/x) * sin(x) - 3*cos(x)/(x*x);
    }
  };
  
  
  class GeneralWindow : public Filter {
    int lobes;
    Filter *subfilter;
    Filter *windowfilter;
  public:
    
    GeneralWindow(int lobes,
		  Filter *filter,
		  Filter *window) {
      this->lobes = lobes;
      subfilter=filter;
      windowfilter=window;
    }
    
    virtual double filter_size(double rate) {
      if (rate > 1.0) rate = 1.0;
      double ret1 = lobes * rate;
      double ret2 = subfilter->filter_size(rate);
      return ret1 < ret2 ? ret1 : ret2;
    }
    
    virtual double filter(double dist) {
      if(fabs(dist) > lobes) return 0.0;
      return subfilter->filter(dist) * windowfilter->filter(dist/lobes);
    }
    
    ~GeneralWindow() {
      delete subfilter;
      delete windowfilter;
    }
  };
  
  // More lobes -> more quality
  Filter *LanczosFilter(int lobes) {
    return new GeneralWindow(lobes, new SincFilter, new SincFilter);
  }

  Matrix<float> create_scaler(int input_size,
			      float begin,
			      float end,
			      int output_size,
			      Filter *filter) {
    Matrix<float> ret(input_size, output_size);
    double rate = output_size / (float)input_size;
    if (rate > 1.0) rate = 1.0;
    float filter_size = filter->filter_size(1.0) / rate;
    for (int o = 0; o < output_size; o++) {
      float opos = (o + 0.5) / output_size * (end - begin) + begin;
      int min_pixel = floor(opos - filter_size);
      int max_pixel = ceil(opos + filter_size);  // floor?
      for (int i = min_pixel; i <= max_pixel; i++) {
	float ipos = i + 0.5;
	float coeff = filter->filter( (ipos - opos) * rate);
	ret.addclamped(i, o, coeff);
      }
    }
    ret.normalizerows();
    return ret;
  }

};  // namespace
