#ifndef SUPERSCALER_TUPLE_H
#define SUPERSCALER_TUPLE_H

#include <assert.h>
#include <math.h>
#include <ostream>

namespace SuperScaler {

// A fixed-size vector containing N values of type T
// T is assumed to have arithmetic operators.
template<class T, int N> class Tuple {
private:
  T data[N];

public:
  // Data access
  const T& operator[](int x) const {
    assert(x>=0);
    assert(x<N);
    return data[x];
  }

  T& operator[](int x) {
    assert(x>=0);
    assert(x<N);
    return data[x];
  }

  Tuple() { for(int e=0;e<N;e++) data[e]=0; }
  Tuple(T x) { for(int e=0;e<N;e++) data[e]=x; }

  // Special for easy 2d use
  Tuple(T a, T b) {
    assert(N == 2);
    data[0]=a;
    data[1]=b;
  }

  // Special for easy triple use
  Tuple(T a, T b, T c) {
    assert(N == 3);
    data[0]=a;
    data[1]=b;
    data[2]=c;
  }

  template<typename FLOAT>
    Tuple(const Tuple<FLOAT, N> &x) {
    for(int e=0;e<N;e++) data[e]=x[e];
  }

  // Assign
  Tuple &operator=(const Tuple &x) {
    if(&x != this)
      for(int e=0;e<N;e++) data[e]=x.data[e];
    return *this;
  }

  // Set every value to x
  Tuple &operator=(const T &x) {
    for(int e=0;e<N;e++) data[e]=x;
    return *this;
  }

  template<typename FLOAT>
  Tuple<FLOAT, N> Real() const {
    Tuple<FLOAT, N> ret(*this);
    return ret;
  }

  Tuple<float, N> Float() const { return Real<float>(); }
  Tuple<double, N> Double() const { return Real<double>(); }

#define BINOP(OP)						\
  Tuple operator OP(const Tuple &x) const {			\
    Tuple ret;							\
    for(int e=0;e<N;e++) ret.data[e]=data[e] OP x.data[e];	\
    return ret;							\
  }								\
  Tuple &operator OP##=(const Tuple &x) {			\
    for(int e=0;e<N;e++) data[e] OP##= x.data[e];		\
    return *this;						\
  }								\
  Tuple operator OP(const T &x) const {				\
    Tuple ret;							\
    for(int e=0;e<N;e++) ret.data[e]=data[e] OP x;		\
    return ret;							\
  }								\
  Tuple &operator OP##=(const T &x) {				\
    for(int e=0;e<N;e++) data[e] OP##= x;			\
    return *this;						\
  }

  BINOP(+)
  BINOP(*)
  BINOP(-)
  BINOP(/)
  BINOP(>>)
  BINOP(<<)

  void addmultiplied(const Tuple &x, T v) {
    for (int i = 0; i < N; i++) data[i] += x[i] * v;
  }

  bool operator==(const T &x) const {
    for(int e=0;e<N;e++) if(!(data[e] == x)) return false;
    return true;
  }

  bool operator==(const Tuple &x) const {
    for(int e=0;e<N;e++) if(!(data[e] == x.data[e])) return false;
    return true;
  }

  // Loop like this:
  // Tuple area_to_iterate_over;
  // Tuple i = area_to_iterate_over.iterator();
  // while (i.next(area_to_iterate_over)) {}
  Tuple iterator() const {
    Tuple ret(0);
    ret.data[0] = -1;
    return ret;
  }

  bool next(const Tuple &upper) {
    for(int e=0;e<N;e++) {
      if(!(++data[e] == upper[e])) return true;
      data[e]=0;
    }
    return false;
  }

  bool next(const Tuple &upper, int exclude) {
    for(int e=0;e<N;e++) {
      if(e == exclude) continue;
      if(!(++data[e] == upper[e])) return true;
      data[e]=0;
    }
    return false;
  }

  T rangeiterator() const {
    T ret = begin();
    ret[0]--;
    return ret;
  }

  bool nextinrange(const Tuple<Tuple, 2> &range) {
    for(int e=0;e<N;e++) {
      if(!(++data[e] == range.end()[e])) return true;
      data[e]=range.begin()[e];
    }
    return false;
  }

  int index(const Tuple &max) const {
    int pos = 0 ;
    for(int e=N-1;e>=0;e--) {
      if (data[e] <0 || data[e] >= max[e]) {
	fprintf(stderr,"data[%d]=%d out of range: [0 .. %d)\n",e,data[e],max[e]);
	assert(false);
      } 
      pos*=max[e];
      pos+=data[e];
    }
    return pos;
  }

  int clampedindex(const Tuple &max) const {
    int pos = 0 ;
    for(int e=N-1; e>=0; e--) {
      pos *= max[e];
      pos += data[e] >= 0 ? data[e] < max[e] ? data[e] : max[e]-1 : 0;
    }
    return pos;
  }

  int clampedmodindex(const Tuple &max, const Tuple &step) const {
    int pos = 0 ;
    for(int e=N-1;e>=0;e--) {
      int x;
      if(data[e] < 0) {
	x=step[e]+~((~data[e]) % step[e]);
      }else{
	if(data[e] >= max[e]) {
	  x = max[e] - max[e] % step[e] + data[e] % step[e];
	  if(x >= max[e]) x-=step[e];
	}else{
	  x=data[e];
	}
      }
      // int x = data[e] >= 0 ? data[e] < max[e] ? data[e] : max[e]-1 : 0;
      assert(x>=0);
      assert(x<max[e]);
      pos*=max[e];
      pos+=x;
    }
    return pos;
  }

  // area index
  int index2(const Tuple<Tuple, 2> &range) {
    return (*this - range.begin()).index(range.Size());
  }

  T Area() const {
    T ret=1;
    for(int e=0;e<N;e++) ret *= data[e];
    return ret;
  }

  // These assume that [0] == beginnnig [1] == end
  // When using a Tuple<T, 2> as a range. 
  const T &begin() const { 
    assert(N==2);
    return data[0];
  }

  const T &end() const { 
    assert(N==2);
    return data[1];
  }

  T &begin() { 
    assert(N==2);
    return data[0];
  }

  T &end() { 
    assert(N==2);
    return data[1];
  }

  T Size() const {
    assert(N==2);
    return end() - begin();
  }

  T UnScale(const T &x) const {
    assert(N == 2);
    return (x - begin()) / Size();
  }
  T Scale(const T &x) const {
    assert(N == 2);
    return x * Size() + begin();
  }

  Tuple MOD(const Tuple &x) const {
    Tuple ret;
    for(int e=0;e<N;e++) {
      ret[e]=(((data[e])<0)?( x[e]+~((~data[e]) % x[e])  ):(data[e])%(x[e]));
//      printf("MOD: %d %% %d = %d\n",data[e],x[e],ret[e]);
    }
    return ret;
  }
};

template<class T, int N> std::ostream& operator<< (std::ostream& out, const Tuple<T, N> &c) {
  out << "[";
  for(int e=0;e<N;e++) {
    if(e) out << ",";
    out << c.data[e];
  }
  out << "]";
  return out;
}

template<int DIM> static inline Tuple<Tuple<int, DIM>, 2> INTAREA(
  const Tuple<Tuple<float, DIM>, 2> &a,
  const Tuple<float, DIM> &b) {
  Tuple<Tuple<int, DIM>, 2> ret;
  for(int e=0;e<DIM;e++) {
    ret[0][e]=(int)floor(a[0][e]-b[e]);
    ret[1][e]=(int)ceil(a[1][e]+b[e]);
  }
  return ret;
}

};

#endif
