#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include "picdb.h"
#include "exif.h"
#include <errno.h>
#include <signal.h>
#include <time.h>

using namespace std;

#define RETR(X) ({ int _; do { _ = (X); } while (_ < 0 && errno == EINTR); _; })

void CreateThread(void *t(void *), void *data) {
  pthread_t tmp;
  pthread_attr_t attr;
  pthread_attr_init(&attr);
  pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
  pthread_create(&tmp, &attr, t, data);
}

void writebytes(int fd, const char *str, size_t len) {
  size_t p = 0;
  while (p < len) {
    ssize_t written = write(fd, str + p, len - p);
    if (written < 0) {
      if (errno == EINTR) continue;
      return;
    }
    p += written;
  }
}

void writestr(int fd, const string &str) {
  fprintf(stderr, "RESPONSE: %s\n", str.c_str());
  writebytes(fd, str.c_str(), str.length());
}

bool endswith(string x, string y) {
  if (x.size() < y.size()) return false;
  return x.rfind(y) == x.size() - y.size();
}

bool ParseRequest(char *data,
		  int bytes,
		  char **path,
		  char **if_modified_since) {
  *path = NULL;
  *if_modified_since = NULL;
  for (int i = bytes - 3; i > 0; i++) {
    if (data[i] == '\n' && data[i+1] == '\r' && data[i+2] == '\n') {
      // Found end of request.
      if (i >= 0  && data[i-1] == '\r') i++;
      data[i] = 0;
      fprintf(stderr, "REQUEST:\n%s\n----------", data);

      if (data[0] != 'G' || data[1] != 'E' || data[2]!= 'T' ||
	  data[3] != ' ') {
	return false;
      }

      // Remove \r
      char *a = data;
      char *b = data;
      while (true) {
	int c = (*b = *(a++));
	if (c == 0) break;
	if (c != '\r') b++;
      }

      char *eol = strchr(data, '\n');
      if (!eol) {
	fprintf(stderr, "Request: Failed to find end of first line!\n");
	return false;
      }
      *eol = 0;
      char *space = strchr(data  + 4, ' ');
      if (space) *space = 0;
      *path = data + 4;

      while (eol - data < i) {
	char *line = eol + 1;
	eol = strchr(line , '\n');
	if (!eol) {
	  fprintf(stderr, "Request: No more lines\n");
	  break;
	} else {
	  *eol = 0;
	  fprintf(stderr, "LINE: %s\n", line);
	}
	char *colon = strchr(line, ':');
	if (!colon) {
	  fprintf(stderr, "Request: No colon on this line, continuing.\n");
	  continue;
	}
	*colon = 0;
	if (!strcasecmp(line, "if-modified-since")) {
	  fprintf(stderr, "Request: if modified since found\n");
	  *if_modified_since = colon + 2;
	}
      }
      return true;
    }
  }
  return false;
}

string timestr(time_t t) {
  char tmp[27];
  ctime_r(&t, tmp);
  char *nl = strrchr(tmp, '\n');
  if (nl) *nl = 0;
  return tmp;
}

string ImgTag(refptr<ImageInterface> &image,
	      const string identifier,
	      const string link,
	      int pixels) {
  int deg = image->rotation();
  string transform = "";
  string divstyle = "";

  char tmp[2048];
  if (deg) {
    pair<int, int> size = image->size();
    sprintf(tmp, "rotate(%ddeg)", deg);
    transform.append(tmp);
    if (deg == 90 || deg == 270 && size.first != 0) {
      int space = pixels - pixels * size.second / size.first;
      if (space < 0) space = 0;
      int translate = space / 2;
      if (deg == 270) translate = - translate;
      sprintf(tmp, " translate(%dpx)", translate);
      transform.append(tmp);
      sprintf(tmp, "width:%dpx; height:%dpx;", pixels, pixels);
      divstyle.append(tmp);
    }
  }
  sprintf(tmp,
	  "<div style='%s'>"
	  "<a href='%s'>"
	  "<img src='%s.jpg' style='-webkit-transform:%s; -ms-transform:%s; transform:%s; max-height: %dpx; max-width: %dpx; width:auto; height:auto;'>"
	  "</a>"
	  "</div>",
	  divstyle.c_str(),
	  link.c_str(),
	  identifier.c_str(),
	  transform.c_str(),
	  transform.c_str(),
	  transform.c_str(),
	  pixels,
	  pixels);
  return tmp;
}

void HandleRequest(int fd) {
  char request[4096];
  int bytes_read = 0;
  char *path;
  char *if_modified_since;

  while (true) {
    int tmp = read(fd, request + bytes_read, 4096 - bytes_read);
    if (tmp < 0 && errno == EINTR) continue;
    if (tmp <= 0) return;
    bytes_read += tmp;
    if (ParseRequest(request, bytes_read, &path, &if_modified_since)) {
      break;
    }
    if (bytes_read >= 4096) {
      fprintf(stderr, "Request too large, bailing!\n");
      return;
    }
  }
  string current_time = timestr(time(NULL));
  fprintf(stderr, "path = %s\n", path);
  string mime_type = "text/html";
  string data;
  char *p = path + 1;
  int pos = -1;
  unsigned long long hash;
  bool got_hash = false;
  if (path[0] == '/' && path[1] == 0) {
    sprintf(request, "HTTP/1.0 303 See Other\r\nLocation: /%d.html\r\n\r\n",
	    GetCurrentHistoryPos());
    writestr(fd, request);
    return;
  }
  if (p[0] >= '0' && p[0] <= '9') {
    pos = strtol(p, &p, 10);
  } else if (p[0] == 'H') {
    got_hash = true;
    hash = strtoull(p+1, &p, 10);
  } else if (p[0] == 'T') {
    // TOP page
    int page = strtol(p+1, &p, 10);
    vector<refptr<ImageInterface> > images;
    GetTopList(page * 20, (page + 1) * 20, &images);
    sprintf(request, "<a href='T%d.html'>Prev Page</a>"
	    "&nbsp;&nbsp;&nbsp;"
	    "<a href='/'>Current Image</a>"
	    "&nbsp;&nbsp;&nbsp;"
	    "<a href='T%d.html'>Next Page</a>\n", page - 1, page + 1);
    string nav(request);
    data.append(nav);
    data.append("<table>\n");
    for (int i = 0; i < images.size(); i++) {
      data.append("<tr><td>");
      sprintf(request, "H%llu", images[i]->hash());
      string identifier(request);
      data.append(ImgTag(images[i], identifier + "_thumb", identifier + ".html", 160));
      unsigned long long h = images[i]->hash();
      sprintf(request, "</td><td>filename: %s<br>score: %d<br></td></tr>\n",
	      images[i]->filename().c_str(),
	      images[i]->score());
      data.append(request);
    }
    data.append("</table>\n");
    data.append(nav);
    data = "HTTP/1.0 200 OK\r\n"
      "Content-Type: text/html\r\n"
      "Date: " + current_time + "\r\n" +
      "Expires: " + current_time + "\r\n" +
      "\r\n" +
      data;
    writestr(fd, data);
    return;
  }
  refptr<ImageInterface> image;
  if (got_hash) {
    image = GetImageByHash(hash);
  } else {
    image = GetHistoryImage(pos);
  } 
  if (!image) {
    writestr(fd, "HTTP/1.0 404 Not found\r\n\r\n");
    return;
  }
  if (p[0] == '?') {
    string action(p + 1);
    if (image) {
      fprintf(stderr, "ACTION=%s\n", action.c_str());
      if (action == "right") image->rotate(90);
      else if (action == "left") image->rotate(-90);
      else if (action == "thumbs_up") image->add_score(1);
      else if (action == "thumbs_down") image->add_score(-1);
      else if (action == "next") {
	refptr<ImageInterface> next_image = GetNextImage(image);
	if (next_image) {
	  hash = next_image->hash();
	  got_hash = true;
	}
      } else if (action == "prev") {
	refptr<ImageInterface> next_image = GetPrevImage(image);
	if (next_image) {
	  hash = next_image->hash();
	  got_hash = true;
	}
      }
    }
    if (got_hash) {
      sprintf(request, "HTTP/1.0 303 See Other\r\nLocation: /H%llu.html\r\n\r\n", hash);
    } else {
      sprintf(request, "HTTP/1.0 303 See Other\r\nLocation: /%d.html\r\n\r\n", pos);
    }
    writestr(fd, request);
    return;
  }
  if (!image) {
    data.append("<body>No current image</body>\n");
  } else {
    string filename = image->filename();
    if (endswith(path, ".jpg")) {
      string last_modified = timestr(GetHistoryTime(pos));

      if (if_modified_since && last_modified == string(if_modified_since)) {
	fprintf(stderr, "MODTIME MATCH!\n");
	writestr(fd, "HTTP/1.0 304 Not Modified\r\n\r\n");
	return;
      }

      data = "HTTP/1.0 200 OK\r\n"
	"Content-Type: image/jpeg\r\n"
	"Last-Modified: " + last_modified + "\r\n"
	"Date: " + current_time + "\r\n"
	"Expires: " + current_time + "\r\n"
	"\r\n";
      writestr(fd, data);
      if (endswith(path, "thumb.jpg")) {
	ExifInterface *exif = GetExifData(filename);
	std::pair<const char *, size_t> thumb_data = exif->GetThumbnail();
	if (thumb_data.first) {
	  writebytes(fd, thumb_data.first, thumb_data.second);
	  delete exif;
	  return;
	}
	delete exif;
      }
      FILE *f = fopen(filename.c_str(), "rb");
      if (f) {
	while (1) {
	  int bytes = fread(request, 1, sizeof(request), f);
	  if (bytes > 0) {
	    writebytes(fd, request, bytes);
	  } else {
	    break;
	  }
	}
	fclose(f);
      }
      return;
    } else {
      string prev, next, identifier;
      if (got_hash) {
	sprintf(request, "H%llu", hash);
	identifier = request;
      } else {
	sprintf(request, "%d", pos);
	identifier = request;
	sprintf(request, "<a href=/%d.html>&lt;</a>", pos - 1);
	prev = request;
	sprintf(request, "<a href=/%d.html>&gt;</a>", pos + 1);
	next = request;
      }
      sprintf(request, "<a href='%s?left'>&#8630;</a>", identifier.c_str());
      string left = request;
      sprintf(request,"<a href='%s?right'>&#8631;</a>", identifier.c_str());
      string right = request;
      sprintf(request, "<a href='%s?thumbs_up'>&#128077;</a>", identifier.c_str());
      string up = request;
      sprintf(request, "<a href='%s?thumbs_down'>&#128078;</a>", identifier.c_str());
      string down = request;

      ExifInterface *exif = GetExifData(image->filename());
      data.append("<head>\n");
      // Turn off underlines for links.
      data.append("<style>\n");
      data.append("a:link {text-decoration:none}\n");
      data.append("a:visited {text-decoration:none}\n");
      data.append("a:hover {text-decoration:none}\n");
      data.append("a:active {text-decoration:none}\n");
      data.append("</style>\n");
      data.append("</head>\n");
      data.append("<body>\n");
      data.append("<center>\n");

      data.append("<span style='font-size:800%' id='topnav'>");
      data.append(prev);
      data.append(left);
      data.append(right);
      data.append(up);
      data.append(down);
      data.append(next);
      data.append("</span><br>");

      data.append("<table><tr><td>");
      data.append("<span style='font-size:800%; display:none;' id='leftnav'><center>");
      data.append(prev);
      data.append("<br>");
      data.append(left);
      data.append("<br>");
      data.append(up);
      data.append("</center></span><br>");
      data.append("</td><td>");
      data.append(ImgTag(image, identifier, "/", 600));
      data.append("</td><td>");
      data.append("<span style='font-size:800% ; display:none;' id='rightnav'><center>");
      data.append(next);
      data.append("<br>");
      data.append(right);
      data.append("<br>");
      data.append(down);
      data.append("</center></span><br>");
      data.append("</td></tr></table>");
      sprintf(request,
	      "<center><a href='%s?prev'>Previous File</a>&nbsp;&nbsp;&nbsp;&nbsp;<a href='%s?next'>Next File</a></center>",
	      identifier.c_str(),
	      identifier.c_str());
      data.append(request);
      data.append("<table>\n");
      set<Filename> filenames;
      image->get_filenames(&filenames);
      time_t file_time = 0;
      for (set<Filename>::const_iterator i = filenames.begin();
	   i != filenames.end();
	   ++i) {
	struct stat s;
	data.append("<tr><td>Filename</td><td>:</td><td>");
	data.append(i->path());
	if (stat(i->path().c_str(), &s) == 0) {
	  if (s.st_mtime > 1000000) {
	    if (s.st_mtime < file_time || file_time == 0) {
	      file_time = s.st_mtime;
	    }
	  }
	}
	data.append("</td></tr>\n");
      }
      
      data.append("<tr><td>Taken</td><td>:</td><td>");
      string date_and_time = exif->describeEntry(EXIF_TAG_DATE_TIME_ORIGINAL);
      if (date_and_time.empty()) {
	date_and_time = exif->describeEntry(EXIF_TAG_DATE_TIME);
      }
      data.append(date_and_time);
      data.append("</td></tr>\n");

      // The oldest file time. (Newer files assumed to be copies.)
      if (file_time) {
	data.append("<tr><td>File time</td><td>:</td><td>");
	data.append(timestr(file_time));
	data.append("</td></tr>\n");
      }

      // Score
      sprintf(request, "%d", image->score());
      data.append("<tr><td>Score</td><td>:</td><td>");
      data.append(request);
      data.append("</td></tr>\n");

      // Rotation
      sprintf(request, "%d", image->rotation());
      data.append("<tr><td>Rotation</td><td>:</td><td>");
      data.append(request);
      data.append("</td></tr>\n");

      data.append("<tr><td>Camera</td><td>:</td><td>");
      data.append(exif->describeEntry(EXIF_TAG_MAKE) + " " +
		  exif->describeEntry(EXIF_TAG_MODEL));
      data.append("</td></tr>\n");

      data.append("<tr><td>Resolution</td><td>:</td><td>");
      data.append(exif->describeEntry(EXIF_TAG_PIXEL_X_DIMENSION) + "x" +
		  exif->describeEntry(EXIF_TAG_PIXEL_Y_DIMENSION));
      data.append("</td></tr>\n");
      delete exif;

      data.append("</table>\n");

      ImageStats stats = GetImageStats();
      sprintf(request, "<br><a href='/T0.html'>%d known pictures, %d upvotes, %d downvotes</a>",
	      stats.num_images,
	      stats.num_upvotes,
	      stats.num_downvotes);
      data.append(request);
	      
      data.append("</center>\n");
      data.append("<script>\n");
      data.append("if (window.innerWidth > window.innerHeight) {\n");
      data.append("   document.getElementById('leftnav').style.display='inline';\n");
      data.append("   document.getElementById('rightnav').style.display='inline';\n");
      data.append("   document.getElementById('topnav').style.display='none';\n");
      data.append("}\n");
      data.append("</script>");
      data.append("</body>\n");
    }
  }

  data = "HTTP/1.0 200 OK\r\n"
    "Content-Type: " + mime_type + "\r\n"
    "Date: " + current_time + "\r\n" +
    "Expires: " + current_time + "\r\n" +
    "\r\n" +
    data;
  writestr(fd, data);
}

void *HandleRequestThread(void *fd) {
  HandleRequest((int)fd);
  close((int)fd);
}

void *RunWebServer(void *unusued) {
  int on = 1;
  int port_fd = socket(AF_INET, SOCK_STREAM, 0);
  RETR(setsockopt(port_fd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on)));
  struct sockaddr_in addr;
  memset(&addr, 0, sizeof(addr));
  addr.sin_family = AF_INET;
  addr.sin_port = htons(80);
  RETR(bind(port_fd, (struct sockaddr*)&addr, sizeof(addr)));
  RETR(listen(port_fd, 5));
  
  while (1) {
    int fd = accept(port_fd, NULL, 0);
    CreateThread(HandleRequestThread, (void *)fd);
  }
}

void StartWebServer() {
  struct sigaction action;
  sigaction(SIGPIPE, NULL, &action);
  action.sa_handler = SIG_IGN;
  sigaction(SIGPIPE, &action, NULL);
  CreateThread(RunWebServer, NULL);
}
